@@ -2,6 +2,13 @@
 # Mount table functions for linux
 
 if (!$no_check_support) {
+	my %support;
+	my $fsfile = &read_file_contents("/proc/filesystems");
+	foreach my $l (split(/\r?\n/, $fsfile)) {
+		my @w = split(/\s+/, $l);
+		my $fs = pop(@w);
+		$support{$fs} = 1;
+		}
 	if (&has_command("amd")) {
 		local $amd = &read_amd_conf();
 		$amd_support = $amd =~ /\[\s*global\s*\]/i ? 2 : 1;
@@ -23,21 +30,22 @@ if (!$no_check_support) {
 		$smbfs_fs = "smbfs";
 		}
 	$swaps_support = -r "/proc/swaps";
-	if (&backquote_command("uname -r") =~ /^(\d+\.\d+)/ && $1 >= 2.4) {
-		$tmpfs_support = 1;
-		$ext3_support = 1;
-		$no_mount_check = 1;
-		$bind_support = 1;	# XXX which version?
-		if ($1 >= 2.6) {
-			$ext4_support = 1;
-			}
+	$tmpfs_support = 1;
+	$ext3_support = 1;
+	$no_mount_check = 1;
+	$bind_support = 1;
+	if ($support{'ext4'}) {
+		$ext4_support = 1;
 		}
-	if (&has_command("mkfs.xfs")) {
+	if ($support{'xfs'} || &has_command("mkfs.xfs")) {
 		$xfs_support = 1;
 		}
-	if (&has_command("mkfs.jfs")) {
+	if ($support{'jfs'} || &has_command("mkfs.jfs")) {
 		$jfs_support = 1;
 		}
+	if ($support{'btrfs'} || &has_command("mkfs.btrfs")) {
+		$btrfs_support = 1;
+		}
 	}
 
 # We always need to check this, to fix up LABEL= mounts
@@ -67,7 +75,7 @@ local(@rv, @p, @o, $_, $i, $j);
 $i = 0;
 
 # Get /etc/fstab mounts
-open(FSTAB, $config{fstab_file});
+open(FSTAB, "<".$config{fstab_file});
 while(<FSTAB>) {
 	local(@o, $at_boot);
 	chop; s/#.*$//g;
@@ -79,6 +87,8 @@ while(<FSTAB>) {
 	else {
 		@p = split(/\s+/, $_);
 		}
+	$p[0] =~ s/\\040/ /g;
+	$p[1] =~ s/\\040/ /g;
 	if ($p[2] eq "proc") { $p[0] = "proc"; }
 	elsif ($p[2] eq "auto") { $p[2] = "*"; }
 	elsif ($p[2] eq "swap") { $p[1] = "swap"; }
@@ -86,11 +96,12 @@ while(<FSTAB>) {
 		$p[0] =~ s/\\040/ /g;
 		$p[0] =~ s/\//\\/g;
 		}
+	$p[1] = &simplify_path($p[1]);
 	$rv[$i] = [ $p[1], $p[0], $p[2] ];
 	$rv[$i]->[5] = "yes";
 	@o = split(/,/ , $p[3] eq "defaults" ? "" : $p[3]);
 	if (($j = &indexof("noauto", @o)) >= 0) {
-		# filesytem is not mounted at boot
+		# filesystem is not mounted at boot
 		splice(@o, $j, 1);
 		$rv[$i]->[5] = "no";
 		}
@@ -140,7 +151,7 @@ elsif ($amd_support == 2) {
 
 # Get kernel automounter configuration
 if ($autofs_support) {
-	open(AUTO, $config{'autofs_file'});
+	open(AUTO, "<".$config{'autofs_file'});
 	while(<AUTO>) {
 		chop;
 		s/#.*$//g;
@@ -162,7 +173,7 @@ return @rv;
 # Add a new entry to the fstab file, old or new automounter file
 sub create_mount
 {
-local(@mlist, @amd, $_); local($opts);
+local(@mlist, @amd, $_, $opts);
 
 if ($_[2] eq "auto") {
 	if ($amd_support == 1) {
@@ -202,10 +213,12 @@ else {
 	local $dev = $_[1];
 	if ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
 		$dev =~ s/\\/\//g;
-		$dev =~ s/ /\\040/g;
 		}
+	$dev =~ s/ /\\040/g;
+	local $mp = $_[0];
+	$mp =~ s/ /\\040/g;
 	&open_tempfile(FSTAB, ">> $config{fstab_file}");
-	&print_tempfile(FSTAB, $dev."\t".$_[0]."\t".$_[2]);
+	&print_tempfile(FSTAB, $dev."\t".$mp."\t".$_[2]);
 	local @opts = $_[3] eq "-" ? ( ) : split(/,/, $_[3]);
 	if ($_[5] eq "no") {
 		push(@opts, "noauto");
@@ -234,7 +247,7 @@ local($i, @fstab, $line, $opts, $j, @amd);
 $i = 0;
 
 # Update fstab file
-open(FSTAB, $config{fstab_file});
+open(FSTAB, "<".$config{fstab_file});
 @fstab = <FSTAB>;
 close(FSTAB);
 &open_tempfile(FSTAB, "> $config{fstab_file}");
@@ -243,8 +256,13 @@ foreach (@fstab) {
 	if ($line =~ /\S/ && $line !~ /\signore\s/ && $i++ == $_[0]) {
 		# Found the line to replace
 		local $dev = $_[2];
+		if ($_[3] eq $smbfs_fs || $_[3] eq "cifs") {
+			$dev =~ s/\\/\//g;
+			}
 		$dev =~ s/ /\\040/g;
-		&print_tempfile(FSTAB, $dev."\t".$_[1]."\t".$_[3]);
+		local $mp = $_[1];
+		$mp =~ s/ /\\040/g;
+		&print_tempfile(FSTAB, $dev."\t".$mp."\t".$_[3]);
 		local @opts = $_[4] eq "-" ? ( ) : split(/,/, $_[4]);
 		if ($_[6] eq "no") {
 			push(@opts, "noauto");
@@ -303,7 +321,7 @@ elsif ($amd_support == 2) {
 
 # Update autofs configuration
 if ($autofs_support) {
-	open(AUTO, $config{'autofs_file'});
+	open(AUTO, "<".$config{'autofs_file'});
 	@auto = <AUTO>;
 	close(AUTO);
 	&open_tempfile(AUTO, "> $config{'autofs_file'}");
@@ -334,7 +352,7 @@ local($i, @fstab, $line, $opts, $j, @amd);
 $i = 0;
 
 # Update fstab file
-open(FSTAB, $config{fstab_file});
+open(FSTAB, "<".$config{fstab_file});
 @fstab = <FSTAB>;
 close(FSTAB);
 &open_tempfile(FSTAB, ">$config{fstab_file}");
@@ -381,7 +399,7 @@ elsif ($amd_support == 2) {
 
 # Update AMD file
 if ($amd_support) {
-	open(AMD, $config{auto_file});
+	open(AMD, "<".$config{auto_file});
 	@amd = <AMD>;
 	close(AMD);
 	&open_tempfile(AMD, ">$config{auto_file}");
@@ -404,7 +422,7 @@ if ($amd_support) {
 
 # Update autofs file
 if ($autofs_support) {
-	open(AUTO, $config{'autofs_file'});
+	open(AUTO, "<".$config{'autofs_file'});
 	@auto = <AUTO>;
 	close(AUTO);
 	&open_tempfile(AUTO, ">$config{'autofs_file'}");
@@ -430,14 +448,29 @@ sub list_mounted
 return @list_mounted_cache
 	if (@list_mounted_cache && $list_mounted_cache_mode == $_[0]);
 local(@rv, @p, @o, $mo, $_, %smbopts);
+
+# Do any filesystems use labels that we need to resolve?
 local @mounts = &list_mounts();
+local $anylabel = 0;
+local $anyuuid = 0;
+foreach my $m (@mounts) {
+	if ($m->[1] =~ /^LABEL=/) {
+		$anylabel++;
+		}
+	if ($m->[1] =~ /^UUID=/) {
+		$anyuuid++;
+		}
+	}
 
 &read_smbopts();
-open(MTAB, "/etc/mtab");
+open(MTAB, "</etc/mtab");
 while(<MTAB>) {
 	chop;
-	s/#.*$//g; if (!/\S/) { next; }
+	s/#.*$//g;
+	if (!/\S/) { next; }
 	@p = split(/\s+/, $_);
+	next if ($p[0] eq "rootfs");	# Bogus pseudo-fs
+	$p[1] = &simplify_path($p[1]);
 	if ($p[2] eq "auto" || $p[0] =~ /^\S+:\(pid\d+\)$/) {
 		# Automounter map.. turn the map= option into the device
 		@o = split(/,/ , $p[3]);
@@ -463,30 +496,13 @@ while(<MTAB>) {
 		# The source for proc mounts is always proc
 		$p[0] = "proc";
 		}
-	if (!$_[0] && ($p[2] =~ /^ext\d+$/ && $has_e2label ||
-	    	       $p[2] eq "xfs" && $has_xfs_db ||
-		       $p[2] eq "reiserfs" && $has_reiserfstune)) {
+	if (!$_[0] && $anylabel) {
 		# Check for a label on this partition, and there is one
 		# and this filesystem is in fstab with the label, change
 		# the device.
-		local $label;
-		if ($p[2] eq "xfs") {
-			local $out = &backquote_command("xfs_db -x -p xfs_admin -c label -r $p[0] 2>&1", 1);
-			$label = $1 if ($out =~ /label\s*=\s*"(.*)"/ &&
-					$1 ne '(null)');
-			}
-		elsif ($p[2] eq "reiserfs") {
-			local $out = &backquote_command("reiserfstune $p[0] 2>&1");
-			if ($out =~ /LABEL:\s*(\S+)/) {
-				$label = $1;
-				}
-			}
-		else {
-			$label = &backquote_command("e2label $p[0] 2>&1", 1);
-			chop($label);
-			}
-		if (!$?) {
-			foreach $m (@mounts) {
+		local $label = &get_filesystem_label(@p);
+		if (defined($label)) {
+			foreach my $m (@mounts) {
 				if ($m->[0] eq $p[1] &&
 				    $m->[1] eq "LABEL=$label") {
 					$p[0] = "LABEL=$label";
@@ -499,7 +515,7 @@ while(<MTAB>) {
 	# Check for a UUID on this partition, and if there is one
 	# and the filesystem is in fstab with the label, change
 	# the device.
-	if (!$_[0]) {
+	if (!$_[0] && $anyuuid) {
 		local $uuid = &device_to_uuid($p[0], \@mounts);
 		if ($uuid) {
 			$p[0] = "UUID=$uuid";
@@ -529,21 +545,21 @@ while(<MTAB>) {
 	push(@rv, [ $p[1], $p[0], $p[2], $p[3] ]);
 	}
 close(MTAB);
-open(SWAPS, "/proc/swaps");
+open(SWAPS, "</proc/swaps");
 while(<SWAPS>) {
 	chop;
 	if (/^(\/\S+)\s+/) {
 		local $sf = $1;
-		if ($sf =~ /^\/dev\/ide\// || $sf =~ /^\/dev\/mapper\//) {
-			# check fstab for a mount on a device which is a symlink
-			local @st = stat($sf);
-			foreach $m (@mounts) {
-				local @fst = stat($m->[1]);
-				if ($m->[2] eq 'swap' && $fst[0] == $st[0] &&
-				    $fst[1] == $st[1]) {
-				    	$sf = $m->[1];
-					last;
-					}
+
+		# check fstab for a mount on a device which is a symlink
+		# to the path in /proc/swaps
+		local @st = stat($sf);
+		foreach $m (@mounts) {
+			local @fst = stat($m->[1]);
+			if ($m->[2] eq 'swap' && $fst[0] == $st[0] &&
+			    $fst[1] == $st[1]) {
+				$sf = $m->[1];
+				last;
 				}
 			}
 
@@ -618,7 +634,8 @@ return $uuid;
 # or an error string if failed
 sub mount_dir
 {
-local($out, $opts, $shar, %options, %smbopts);
+local($out, $opts, $shar, %options, %smbopts, $qdir);
+$qdir = quotemeta($_[0]);
 local @opts = $_[3] eq "-" || $_[3] eq "" ? ( ) :
 		grep { $_ ne "noauto" } split(/,/, $_[3]);
 if ($_[2] eq "bind") {
@@ -653,7 +670,8 @@ if ($_[2] eq "swap") {
 	}
 elsif ($_[2] eq "auto") {
 	# Old automounter filesystem
-	$out = &backquote_logged("amd $_[0] $_[1] >/dev/null 2>/dev/null");
+	$out = &backquote_logged(
+		"amd $qdir ".quotemeta($_[1])." >/dev/null 2>/dev/null");
 	if ($?) { return $text{'linux_eamd'}; }
 	}
 elsif ($_[2] eq "autofs") {
@@ -661,7 +679,8 @@ elsif ($_[2] eq "autofs") {
 	$opts = &autofs_args($_[3]);
 	$type = $_[1] !~ /^\// ? "yp" :
 		(-x $_[1]) ? "program" : "file";
-	$out = &backquote_logged("automount $opts $_[0] $type $_[1] 2>&1");
+	local $qmap = quotemeta($_[1]);
+	$out = &backquote_logged("automount $opts $qdir $type $qmap 2>&1");
 	if ($?) { return &text('linux_eauto', "<pre>$out</pre>"); }
 	}
 elsif ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
@@ -673,29 +692,30 @@ elsif ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
 	if ($support >= 3) {
 		# SMB filesystem mounted with mount command
 		local $temp = &transname();
-		local $ex = &system_logged("mount -t $_[2] $opts $qshar $_[0] >$temp 2>&1 </dev/null");
-		local $out = `cat $temp`;
+		local $qtype = quotemeta($_[2]);
+		local $ex = &system_logged("mount -t $qtype $opts $qshar $qdir >$temp 2>&1 </dev/null");
+		local $out = &read_file_contents($temp);
 		unlink($temp);
 		if ($ex || $out =~ /failed|error/i) {
-			&system_logged("umount $_[0] >/dev/null 2>&1");
+			&system_logged("umount $qdir >/dev/null 2>&1");
 			return "<pre>$out</pre>";
 			}
 		}
 	elsif ($support == 2) {
 		# SMB filesystem mounted with version 2.x smbmount
 		$opts =
-		    ($options{'user'} ? "-U $options{'user'} " : "").
+		    ($options{'user'} ? "-U ".quotemeta($options{'user'})." " : "").
 		    ($options{'passwd'} ? "" : "-N ").
-		    ($options{'workgroup'} ? "-W $options{'workgroup'} " : "").
-		    ($options{'clientname'} ? "-n $options{'clientname'} " : "").
-		    ($options{'machinename'} ? "-I $options{'machinename'} " : "");
+		    ($options{'workgroup'} ? "-W ".quotemeta($options{'workgroup'})." " : "").
+		    ($options{'clientname'} ? "-n ".quotemeta($options{'clientname'})." " : "").
+		    ($options{'machinename'} ? "-I ".quotemeta($options{'machinename'})." " : "");
 		&foreign_require("proc");
 		local ($fh, $fpid) = &proc::pty_process_exec_logged(
-			"sh -c 'smbmount $shar $_[0] -d 0 $opts'");
+			"sh -c 'smbmount $qshar $qdir -d 0 $opts'");
 		if ($options{'passwd'}) {
 			local $w = &wait_for($fh, "word:");
 			if ($w < 0) {
-				&system_logged("umount $_[0] >/dev/null 2>&1");
+				&system_logged("umount $qdir >/dev/null 2>&1");
 				return $text{'linux_esmbconn'};
 				}
 			local $p = "$options{'passwd'}\n";
@@ -706,7 +726,7 @@ elsif ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
 			$got .= $_;
 			}
 		if ($got =~ /failed/) {
-			&system_logged("umount $_[0] >/dev/null 2>&1");
+			&system_logged("umount $qdir >/dev/null 2>&1");
 			return "<pre>$got</pre>\n";
 			}
 		close($fh);
@@ -716,19 +736,28 @@ elsif ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
 		$shortname = &get_system_hostname();
 		if ($shortname =~ /^([^\.]+)\.(.+)$/) { $shortname = $1; }
 		$opts =
-		   ($options{servername} ? "-s $options{servername} " : "").
-		   ($options{clientname} ? "-c $options{clientname} "
-					 : "-c $shortname ").
-		   ($options{machinename} ? "-I $options{machinename} " : "").
-		   ($options{user} ? "-U $options{user} " : "").
-		   ($options{passwd} ? "-P $options{passwd} " : "-n ").
-		   ($options{uid} ? "-u $options{uid} " : "").
-		   ($options{gid} ? "-g $options{gid} " : "").
-		   ($options{fmode} ? "-f $options{fmode} " : "").
-		   ($options{dmode} ? "-d $options{dmode} " : "");
-		$out = &backquote_logged("smbmount $shar $_[0] $opts 2>&1 </dev/null");
+		   ($options{servername}
+		   	? "-s ".quotemeta($options{servername})." " : "").
+		   ($options{clientname}
+		   	? "-c ".quotemeta($options{clientname})." "
+			: "-c ".quotemeta($shortname)." ").
+		   ($options{machinename}
+		   	? "-I ".quotemeta($options{machinename})." " : "").
+		   ($options{user}
+		   	? "-U ".quotemeta($options{user})." " : "").
+		   ($options{passwd}
+		   	? "-P ".quotemeta($options{passwd})." " : "-n ").
+		   ($options{uid}
+		   	? "-u ".quotemeta($options{uid})." " : "").
+		   ($options{gid}
+		   	? "-g ".quotemeta($options{gid})." " : "").
+		   ($options{fmode}
+		   	? "-f ".quotemeta($options{fmode})." " : "").
+		   ($options{dmode}
+		   	? "-d ".quotemeta($options{dmode})." " : "");
+		$out = &backquote_logged("smbmount $qshar $qdir $opts 2>&1 </dev/null");
 		if ($out) {
-			&system_logged("umount $_[0] >/dev/null 2>&1");
+			&system_logged("umount $qdir >/dev/null 2>&1");
 			return "<pre>$out</pre>";
 			}
 		}
@@ -739,7 +768,8 @@ elsif ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
 else {
 	# some filesystem supported by mount
 	local $fs = $_[2] eq "*" ? "auto" : $_[2];
-	$cmd = "mount -t $fs $opts $devargs ".quotemeta($_[0]);
+	local $qfs = quotemeta($fs);
+	$cmd = "mount -t $qfs $opts $devargs $qdir";
 	$out = &backquote_logged("$cmd 2>&1 </dev/null");
 	if ($?) { return "<pre>$out</pre>"; }
 	}
@@ -765,14 +795,19 @@ else {
 	push(@opts, "remount");
 	local $opts = @opts ? "-o ".quotemeta(join(",", @opts)) : "";
 	local $fs = $_[2] eq "*" ? "auto" : $_[2];
+	local $qfs = quotemeta($fs);
+	local $qdir = quotemeta($_[0]);
+	local $qdev = quotemeta($_[1]);
 	if ($_[1] =~ /LABEL=(.*)/) {
-		$cmd = "mount -t $fs -L $1 $opts $_[0]";
+		local $qlabel = quotemeta($1);
+		$cmd = "mount -t $qfs -L $qlabel $opts $qdir";
 		}
 	elsif ($_[1] =~ /UUID=(\S+)/) {
-		$cmd = "mount -t $fs -U $1 $opts $_[0]";
+		local $quuid = quotemeta($1);
+		$cmd = "mount -t $qfs -U $quuid $opts $qdir";
 		}
 	else {
-		$cmd = "mount -t $fs $opts $_[1] $_[0]";
+		$cmd = "mount -t $qfs $opts $qdev $qdir";
 		}
 	$out = &backquote_logged("$cmd 2>&1 </dev/null");
 	if ($?) { return "<pre>$out</pre>"; }
@@ -786,10 +821,12 @@ else {
 # or an error string if failed
 sub unmount_dir
 {
-local($out, %smbopts, $dir);
+local($out, %smbopts, $dir, $qdir, $qdev);
+$qdir = quotemeta($_[0]);
+$qdev = quotemeta($_[1]);
 if ($_[2] eq "swap") {
 	# Use swapoff to remove the swap space..
-	$out = &backquote_logged("swapoff $_[1]");
+	$out = &backquote_logged("swapoff $qdev");
 	if ($?) { return "<pre>$out</pre>"; }
 	}
 elsif ($_[2] eq "auto") {
@@ -809,7 +846,7 @@ elsif ($_[2] eq "autofs") {
 	}
 else {
 	local $fflag = $_[4] ? "-l" : "";
-	$out = &backquote_logged("umount $fflag $_[0] 2>&1");
+	$out = &backquote_logged("umount $fflag $qdir 2>&1");
 	if ($?) { return "<pre>$out</pre>"; }
 	if ($_[2] eq $smbfs_fs || $_[2] eq "cifs") {
 		# remove options from list
@@ -871,7 +908,7 @@ elsif ($_[0] eq $smbfs_fs)
 	{ return ($smbfs_support >= 3 ? 2 : 0, 1, 0, 0); }
 elsif ($_[0] eq "cifs") { return (2, 1, 0, 0); }
 elsif ($_[0] =~ /^ext\d+$/ || $_[0] eq "minix" ||
-       $_[0] eq "xiafs" || $_[0] eq "xfs" || $_[0] eq "jfs")
+       $_[0] eq "xiafs" || $_[0] eq "xfs" || $_[0] eq "jfs" || $_[0] eq "btrfs")
 	{ return (2, 1, 1, 0); }
 else
 	{ return (2, 1, 0, 0); }
@@ -889,8 +926,8 @@ if ($_[0] eq "proc" || $_[0] eq "swap" ||
 &clean_language();
 local $out = &backquote_command("df -k ".quotemeta($_[1]), 1);
 &reset_environment();
-if ($out =~ /Mounted on\n\S+\s+(\S+)\s+\S+\s+(\S+)/) {
-	return ($1, $2);
+if ($out =~ /Mounted on\n\S+\s+(?<total>\S+)\s+(?<used>\S+)\s+(?<free>\S+)\s+(?<percent>\d+)/) {
+	return ("$+{total}", "$+{free}", "$+{used}", "$+{percent}");
 	}
 return ( );
 }
@@ -903,8 +940,8 @@ if (&get_mounted($_[1], "*") < 0) { return (); }
 &clean_language();
 local $out = &backquote_command("df -i $_[1]", 1);
 &reset_environment();
-if ($out =~ /Mounted on\n\S+\s+(\S+)\s+\S+\s+(\S+)/) {
-	return ($1, $2);
+if ($out =~ /Mounted on\n\S+\s+(?<total>\S+)\s+(?<used>\S+)\s+(?<free>\S+)\s+(?<percent>\d+)/) {
+	return ("$+{total}", "$+{free}", "$+{used}", "$+{percent}");
 	}
 return ( );
 }
@@ -915,8 +952,8 @@ return ( );
 # generate_options() will not be called for it.
 sub list_fstypes
 {
-local @sup = ("ext2", "minix", "msdos", "nfs", "nfs4", "iso9660", "ext",
-	      "xiafs", "hpfs", "fat", "vfat", "umsdos", "sysv", "reiserfs",
+local @sup = ("ext2", "msdos", "nfs", "nfs4", "iso9660", "ext",
+	      "hpfs", "fat", "vfat", "umsdos", "sysv", "reiserfs",
 	      "ntfs", "hfs", "fatx");
 push(@sup, $smbfs_fs) if ($smbfs_support);
 push(@sup, "cifs") if ($cifs_support);
@@ -926,12 +963,20 @@ push(@sup, "tmpfs") if ($tmpfs_support);
 push(@sup, "ext3") if ($ext3_support);
 push(@sup, "ext4") if ($ext4_support);
 push(@sup, "xfs") if ($xfs_support);
+push(@sup, "btrfs") if ($btrfs_support);
 push(@sup, "jfs") if ($jfs_support);
 push(@sup, "bind") if ($bind_support);
 push(@sup, "swap");
 return @sup;
 }
 
+# preferred_fstype()
+# Returns the default filesystem type for this OS
+sub preferred_fstype
+{
+return $ext4_support ? "ext4" :
+       $ext3_support ? "ext3" : "ext2";
+}
 
 # fstype_name(type)
 # Given a short filesystem type, return a human-readable name for it
@@ -970,12 +1015,13 @@ local(%fsmap);
 	  "btrfs","Oracle B-Tree Filesystem",
 	  "ocfs2","Oracle Clustering Filesystem",
 	  "gfs2","RedHat Clustering Filesystem",
-	  "xfs","SGI Filesystem",
+	  "xfs","Linux XFS Filesystem",
 	  "jfs","IBM Journalling Filesystem",
 	  "ntfs","Windows NT Filesystem",
 	  "bind","Loopback Filesystem",
 	  "hfs","Apple Filesystem",
 	  "fatx","XBOX Filesystem",
+	  "btrfs","BTR Filesystem",
 	  );
 return $config{long_fstypes} && $fsmap{$_[0]} ? $fsmap{$_[0]} : uc($_[0]);
 }
@@ -1087,9 +1133,9 @@ else {
 		local @ropts;
 		foreach $c (@$conf) {
 			if ($c->{'active'}) {
+				$c->{'value'} =~ /(\d+)$/;
 				push(@ropts, [ $c->{'value'},
-					&text('linux_rdev',
-					      substr($c->{'value'}, -1)) ]);
+					       &text('linux_rdev', "$1") ]);
 				$rfound++ if ($loc eq $c->{'value'});
 				}
 			}
@@ -1197,8 +1243,11 @@ if ($type ne "swap" && $type ne "auto" &&
 	print &ui_table_row(&hlink($text{'linux_noatime'}, "linux_noatime"),
 		&ui_yesno_radio("lnx_noatime", defined($options{"noatime"})));
 
-	print &ui_table_row($text{'linux_netdev'},
+	print &ui_table_row(&hlink($text{'linux_netdev'}, "linux_netdev"),
 		&ui_yesno_radio("lnx_netdev", defined($options{"_netdev"})));
+
+	print &ui_table_row(&hlink($text{'linux_nofail'}, "linux_nofail"),
+		&ui_yesno_radio("lnx_nofail", defined($options{"nofail"})));
 	}
 	
 if ($type =~ /^ext\d+$/) {
@@ -1233,6 +1282,10 @@ if ($type =~ /^ext\d+$/) {
 		&ui_yesno_radio("ext2_grpid", defined($options{"grpid"}) ||
 					      defined($options{"bsdgroups"})));
 
+	my $usrquota = defined($options{"usrquota"}) ||
+			defined($options{"usrjquota"});
+	my $grpquota = defined($options{"grpquota"}) ||
+			defined($options{"grpjquota"});
 	print &ui_table_row($text{'linux_quotas'},
 		&ui_select("ext2_quota", $usrquota && $grpquota ? 3 :
 					 $grpquota ? 2 :
@@ -1242,13 +1295,20 @@ if ($type =~ /^ext\d+$/) {
 			     [ 2, $text{'linux_grpquota'} ],
 			     [ 3, $text{'linux_usrgrpquota'} ] ]));
 
+	print &ui_table_row($text{'linux_quotaj'},
+		&ui_radio("ext2_quotaj",
+			  defined($options{"usrjquota"}) ||
+			   defined($options{"grpjquota"}) ? 1 : 0,
+			  [ [ 1, $text{'linux_quotaj1'} ],
+			    [ 0, $text{'linux_quotaj0'} ] ]));
+
 	print &ui_table_row($text{'linux_resuid'},
 		&ui_user_textbox("ext2_resuid", defined($options{"resuid"}) ?
 				   getpwuid($options{"resuid"}) : ""));
 
 	print &ui_table_row($text{'linux_resgid'},
 		&ui_group_textbox("ext2_resgid", defined($options{"resgid"}) ?
-				   getpwgid($options{"resgid"}) : ""));
+				   getgrgid($options{"resgid"}) : ""));
 	}
 elsif ($type eq "nfs" || $type eq "nfs4") {
 	# Linux nfs has some more options...
@@ -1291,19 +1351,27 @@ elsif ($type eq "nfs" || $type eq "nfs4") {
 		&ui_opt_textbox("nfs_wsize", $options{"wsize"}, 6,
 				$text{'default'}));
 
+	my ($auth, $sec) = $options{"sec"} =~ /^(ntlmssp|ntlmv2|ntlm|spkm|lkey|krb5|sys)(p|i|)/ ? ($1, $2) : ( );
 	print &ui_table_row(&hlink($text{'linux_auth'}, "linux_auth"),
-		&ui_radio("nfs_auth", $options{"sec"} =~ /spkm/ ? 3 :
-				      $options{"sec"} =~ /lipkey/ ? 2 :
-				      $options{"sec"} =~ /krb5/ ? 1 : 0,
-			  [ [ 0, 'sys' ], [ 1, 'krb5' ],
-			    [ 2, 'lipkey' ], [ 3, 'spkm-3' ] ]));
+		&ui_radio("nfs_auth", $auth,
+			  [ [ '', 'sys' ],
+			    [ 'krb5', 'krb5 (Kerberos 5)' ],
+			    [ 'lkey', 'lkey' ],
+			    [ 'spkm', 'spkm-3' ],
+			    [ 'ntlm', 'ntlm (NTLM)' ],
+			    [ 'ntmlv2', 'ntmlv2 (NTLM version 2)' ],
+			    [ 'ntlmssp', 'ntlmssp (NTLMv2 in NTLMSSP)' ] ]));
 
 	print &ui_table_row(&hlink($text{'linux_sec'}, "linux_sec"),
-		&ui_radio("nfs_sec", $options{"sec"} =~ /i$/ ? 1 :
-				     $options{"sec"} =~ /p$/ ? 2 : 0,
-			  [ [ 0, $text{'config_none'} ],
-			    [ 1, $text{'linux_integrity'} ],
-			    [ 2, $text{'linux_privacy'} ] ]));
+		&ui_radio("nfs_sec", $sec,
+			  [ [ '', $text{'config_none'} ],
+			    [ 'i', $text{'linux_integrity'} ],
+			    [ 'p', $text{'linux_privacy'} ] ]));
+
+	print &ui_table_row(&hlink($text{'linux_nfsvers'}, "linux_nfsvers"),
+		&ui_select("nfs_nfsvers", $options{"nfsvers"},
+			   [ [ "", $text{'linux_nfsdefault'} ],
+			     2, 3, 4, 4.1 ]));
 	}
 elsif ($type eq "fat" || $type eq "vfat" || $type eq "msdos" ||
        $type eq "umsdos" || $type eq "fatx"){
@@ -1484,11 +1552,13 @@ elsif ($type eq $smbfs_fs || $type eq "cifs") {
 		}
 	elsif ($support >= 3) {
 		print &ui_table_row($text{'linux_fmode'},
-			&ui_opt_textbox("smbfs_fmask", $options{'fmask'}, 5,
+			&ui_opt_textbox("smbfs_file_mode",
+					$options{'file_mode'}, 5,
 					$text{'default'}));
 
 		print &ui_table_row($text{'linux_dmode'},
-			&ui_opt_textbox("smbfs_dmask", $options{'dmask'}, 5,
+			&ui_opt_textbox("smbfs_dir_mode",
+					$options{'dir_mode'}, 5,
 					$text{'default'}));
 
 		print &ui_table_row($text{'linux_ro'},
@@ -1521,6 +1591,18 @@ elsif ($type eq $smbfs_fs || $type eq "cifs") {
 		print &ui_table_row($text{'linux_iocharset'},
 			&ui_opt_textbox("smbfs_iocharset",
 			    $options{'iocharset'}, 10, $text{'default'}));
+
+		print &ui_table_row($text{'linux_nounix'},
+			&ui_yesno_radio("smbfs_nounix",
+					defined($options{"nounix"})));
+
+		print &ui_table_row($text{'linux_cvers'},
+			&ui_opt_textbox("smbfs_vers", $options{"vers"},
+					5, $text{'linux_auto'}));
+
+		print &ui_table_row($text{'linux_noserverino'},
+			&ui_yesno_radio("smbfs_noserverino",
+					defined($options{"noserverino"})));
 		}
 	}
 elsif ($type eq "reiserfs") {
@@ -1596,21 +1678,23 @@ elsif ($type eq "ntfs") {
 sub check_location
 {
 if (($_[0] eq "nfs") || ($_[0] eq "nfs4")) {
-	local($out, $temp, $mout, $dirlist);
+	local($out, $temp, $mout, $dirlist, @dirlist);
 
-	if (&has_command("showmount")) {
+	if (&has_command("showmount") && $config{'nfs_check'}) {
 		# Use ping and showmount to see if the host exists and is up
 		if ($in{nfs_host} !~ /^\S+$/) {
 			&error(&text('linux_ehost', $in{'nfs_host'}));
 			}
-		$out = &backquote_command("ping -c 1 '$in{nfs_host}' 2>&1");
+		$out = &backquote_command(
+			"ping -c 1 ".quotemeta($in{nfs_host})." 2>&1");
 		if ($out =~ /unknown host/) {
 			&error(&text('linux_ehost2', $in{'nfs_host'}));
 			}
 		elsif ($out =~ /100\% packet loss/) {
 			&error(&text('linux_edown', $in{'nfs_host'}));
 			}
-		$out = &backquote_command("showmount -e '$in{nfs_host}' 2>&1");
+		$out = &backquote_command(
+			"showmount -e ".quotemeta($in{nfs_host})." 2>&1");
 		if ($out =~ /Unable to receive/) {
 			&error(&text('linux_enfs', $in{'nfs_host'}));
 			}
@@ -1620,39 +1704,18 @@ if (($_[0] eq "nfs") || ($_[0] eq "nfs4")) {
 
 		# Validate directory name for NFSv3 (in v4 '/' exists)
 		foreach (split(/\n/, $out)) {
-			if (/^(\/\S+)/) { $dirlist .= "$1\n"; }
+			if (/^(\/\S+)/) {
+				$dirlist .= "$1\n";
+				push(@dirlist, $1);
+				}
 			}
-		
-		if (($_[0] ne "nfs4") && ($in{nfs_dir} !~ /^\/.*$/)) {
+	
+		if ($_[0] ne "nfs4" && $in{'nfs_dir'} !~ /^\/.*$/ &&
+		    &indexof($in{'nfs_dir'}, @dirlist) < 0) {
 			&error(&text('linux_enfsdir', $in{'nfs_dir'},
 				     $in{'nfs_host'}, "<pre>$dirlist</pre>"));
-		    }
-	    }
-
-	# Try a test mount to see if filesystem is available
-	$temp = &tempname();
-	&make_dir($temp, 0755);
-	&execute_command("mount -t $_[0] ".
-			 quotemeta("$in{nfs_host}:$in{nfs_dir}")." ".
-			 quotemeta($temp),
-			 undef, \$mout, \$mout);
-	if ($mout =~ /No such file or directory/i) {
-		&error(&text('linux_enfsdir', $in{'nfs_dir'},
-			     $in{'nfs_host'}, "<pre>$dirlist</pre>"));
+			}
 		}
-	elsif ($mout =~ /Permission denied/i) {
-		&error(&text('linux_enfsperm', $in{'nfs_dir'}, $in{'nfs_host'}));
-		}
-	elsif ($?) {
-		&error(&text('linux_enfsmount', "<tt>$mout</tt>"));
-		}
-	# It worked! unmount
-	local $umout;
-	&execute_command("umount ".quotemeta($temp), undef, \$umout, \$umout);
-	if ($?) {
-		&error(&text('linux_enfsmount', "<tt>$umout</tt>"));
-		}
-	rmdir(&translate_filename($temp));	# Don't delete mounted files!
 
 	return "$in{nfs_host}:$in{nfs_dir}";
 	}
@@ -1765,6 +1828,9 @@ if ($_[0] ne "swap" && $_[0] ne "auto" &&
 
 	delete($options{"_netdev"});
 	$options{"_netdev"} = "" if ($in{'lnx_netdev'});
+
+	delete($options{"nofail"});
+	$options{"nofail"} = "" if ($in{'lnx_nofail'});
 	}
 
 if (($_[0] eq "nfs") || ($_[0] eq "nfs4")) {
@@ -1776,7 +1842,11 @@ if (($_[0] eq "nfs") || ($_[0] eq "nfs4")) {
 	if ($in{nfs_soft}) { $options{"soft"} = ""; }
 
 	delete($options{"timeo"});
-	if (!$in{nfs_timeo_def}) { $options{"timeo"} = $in{nfs_timeo}; }
+	if (!$in{nfs_timeo_def}) {
+		$in{nfs_timeo} =~ /^\d+$/ && $in{nfs_timeo} > 0 ||
+			&error($text{'linux_etimeo'});
+		$options{"timeo"} = $in{nfs_timeo};
+		}
 
 	delete($options{"retrans"});
 	if (!$in{nfs_retrans_def}) { $options{"retrans"} = $in{nfs_retrans}; }
@@ -1802,11 +1872,16 @@ if (($_[0] eq "nfs") || ($_[0] eq "nfs4")) {
 	if (!$in{nfs_rsize_def}) { $options{"rsize"} = $in{nfs_rsize}; }
 
 	delete($options{"sec"});
+
 	# Only sys and krb5 for the moment
 	if ($in{nfs_auth}) {
-	    if ($in{nfs_sec} == 0) { $options{"sec"} = "krb5"; }
-	    if ($in{nfs_sec} == 1) { $options{"sec"} = "krb5i"; }
-	    if ($in{nfs_sec} == 2) { $options{"sec"} = "krb5p"; }
+	    $options{"sec"} = $in{"nfs_auth"}.$in{"nfs_sec"};
+	}
+
+	if ($in{'nfs_nfsvers'}) {
+		$options{'nfsvers'} = $in{'nfs_nfsvers'};
+	} else {
+		delete($options{'nfsvers'});
 	}
     }
 elsif ($_[0] =~ /^ext\d+$/) {
@@ -1839,12 +1914,26 @@ elsif ($_[0] =~ /^ext\d+$/) {
 	if ($in{'ext2_resgid'})
 		{ $options{"resgid"} = getgrnam($in{'ext2_resgid'}); }
 
+	my $jufile = $options{"usrjquota"};
+	my $jgfile = $options{"grpjquota"};
 	delete($options{"quota"}); delete($options{"noquota"});
 	delete($options{"usrquota"}); delete($options{"grpquota"});
-	if ($in{'ext2_quota'} == 1) { $options{'usrquota'} = ""; }
-	elsif ($in{'ext2_quota'} == 2) { $options{'grpquota'} = ""; }
-	elsif ($in{'ext2_quota'} == 3)
-		{ $options{'usrquota'} = $options{'grpquota'} = ""; }
+	delete($options{"usrjquota"}); delete($options{"grpjquota"});
+	my ($u, $g) = ("usrquota", "grpquota");
+	if ($in{'ext2_quotaj'}) {
+		($u, $g) = ("usrjquota", "grpjquota");
+		$jufile ||= "aquota.user";
+		$jgfile ||= "aquota.group";
+		$options{"jqfmt"} = "vfsv0";
+		}
+	else {
+		$jufile = "";
+		$jgfile = "";
+		delete($options{"jqfmt"});
+		}
+	if ($in{'ext2_quota'} == 1) { $options{$u} = $jufile; }
+	elsif ($in{'ext2_quota'} == 2) { $options{$g} = $jgfile; }
+	elsif ($in{'ext2_quota'} == 3) { $options{$u} = $jufile; $options{$g} = $jgfile; }
 	}
 elsif ($_[0] eq "fat" || $_[0] eq "vfat" ||
        $_[0] eq "msdos" || $_[0] eq "umsdos" || $_[0] eq "fatx") {
@@ -1978,7 +2067,9 @@ elsif ($_[0] eq $smbfs_fs || $_[0] eq "cifs") {
 			# No hostname found for the server.. try to guess
 			local($out, $sname);
 			$sname = $in{'smbfs_server'};
-			$out = &backquote_command("$config{'nmblookup_path'} -d 0 $sname 2>&1");
+			local $qsname = quotemeta($sname);
+			$out = &backquote_command(
+				"$config{'nmblookup_path'} -d 0 $qsname 2>&1");
 			if (!$? && $out =~ /^([0-9\.]+)\s+$sname\n/) {
 				$options{machinename} = $1;
 				}
@@ -2015,22 +2106,22 @@ elsif ($_[0] eq $smbfs_fs || $_[0] eq "cifs") {
 			{ $options{dmode} = $in{smbfs_dmode}; }
 		}
 	elsif ($support >= 3) {
-		if ($in{'smbfs_fmask_def'}) {
-			delete($options{'fmask'});
+		if ($in{'smbfs_file_mode_def'}) {
+			delete($options{'file_mode'});
 			}
 		else {
-			$in{'smbfs_fmask'} =~ /^[0-7]{3}$/ ||
-			    &error(&text('linux_efmode', $in{'smbfs_fmask'}));
-			$options{'fmask'} = $in{'smbfs_fmask'};
+			$in{'smbfs_file_mode'} =~ /^0?[0-7]{3}$/ ||
+			  &error(&text('linux_efmode', $in{'smbfs_file_mode'}));
+			$options{'file_mode'} = $in{'smbfs_file_mode'};
 			}
 
-		if ($in{'smbfs_dmask_def'}) {
-			delete($options{'dmask'});
+		if ($in{'smbfs_dir_mode_def'}) {
+			delete($options{'dir_mode'});
 			}
 		else {
-			$in{'smbfs_dmask'} =~ /^[0-7]{3}$/ ||
-			    &error(&text('linux_edmode', $in{'smbfs_dmask'}));
-			$options{'dmask'} = $in{'smbfs_dmask'};
+			$in{'smbfs_dir_mode'} =~ /^0?[0-7]{3}$/ ||
+		    	  &error(&text('linux_edmode', $in{'smbfs_dir_mode'}));
+			$options{'dir_mode'} = $in{'smbfs_dir_mode'};
 			}
 
 		delete($options{'ro'}); delete($options{'rw'});
@@ -2056,6 +2147,19 @@ elsif ($_[0] eq $smbfs_fs || $_[0] eq "cifs") {
 				&error($text{'linux_eiocharset'});
 			$options{'iocharset'} = $in{'smbfs_iocharset'};
 			}
+
+		delete($options{'nounix'});
+		if ($in{'smbfs_nounix'}) { $options{'nounix'} = ''; }
+
+		delete($options{'noserverino'});
+		if ($in{'smbfs_noserverino'}) { $options{'noserverino'} = ''; }
+
+		delete($options{'vers'});
+		if (!$in{'smbfs_vers_def'}) {
+			$in{'smbfs_vers'} =~ /^[0-9\.]+$/ ||
+				&error($text{'linux_ecvers'});
+			$options{'vers'} = $in{'smbfs_vers'};
+			}
 		}
 	}
 elsif ($_[0] eq "reiserfs") {
@@ -2169,7 +2273,7 @@ return @rv ? join(',' , @rv) : "-";
 sub read_smbopts
 {
 local($_);
-open(SMBOPTS, "$module_config_directory/smbfs");
+open(SMBOPTS, "<$module_config_directory/smbfs");
 while(<SMBOPTS>) {
 	/^(\S+)\s+(\S+)$/;
 	$smbopts{$1} = $2;
@@ -2192,13 +2296,14 @@ foreach (keys %smbopts) {
 # Calls dd and mkswap to setup a swap file
 sub create_swap
 {
-local($out, $bl);
+local($out, $bl, $qfile);
+$qfile = quotemeta($_[0]);
 $bl = $_[1] * ($_[2] eq "t" ? 1024*1024*1024 :
 	       $_[2] eq "g" ? 1024*1024 :
 	       $_[2] eq "m" ? 1024 : 1);
-$out = &backquote_logged("dd if=/dev/zero of=$_[0] bs=1024 count=$bl 2>&1");
+$out = &backquote_logged("dd if=/dev/zero of=$qfile bs=1024 count=$bl 2>&1");
 if ($?) { return "dd failed : $out"; }
-$out = &backquote_logged("mkswap $_[0] $bl 2>&1");
+$out = &backquote_logged("mkswap $qfile $bl 2>&1");
 if ($?) { return "mkswap failed : $out"; }
 &system_logged("sync >/dev/null 2>&1");
 return 0;
@@ -2288,7 +2393,7 @@ local $sl = $/;
 $/ = undef;
 local $rv;
 foreach $f (split(/\s+/, $config{'auto_file'})) {
-	open(AMD, $f);
+	open(AMD, "<".$f);
 	$rv .= <AMD>;
 	close(AMD);
 	}
@@ -2312,7 +2417,7 @@ sub parse_amd_conf
 local (@rv, $str);
 foreach $f (split(/\s+/, $config{'auto_file'})) {
 	local $lnum = 0;
-	open(AMD, $f);
+	open(AMD, "<".$f);
 	while(<AMD>) {
 		s/\r|\n//g;
 		s/#.*$//g;
@@ -2369,6 +2474,10 @@ return $_[0] =~ /^\/dev\/(s|h|xv|v)d([a-z]+)(\d+)$/ ?
 	&text('select_device', $1 eq 's' ? 'SCSI' : $1 eq 'xv' ? 'Xen' :
 			       $1 eq 'v' ? 'VirtIO' : 'IDE',
 			       uc($2)) :
+       $_[0] =~ /^\/dev\/mmcblk(\d+)p(\d+)$/ ?
+	&text('select_part', "SD-Card", "$1", "$2") :
+       $_[0] =~ /^\/dev\/mmcblk(\d+)$/ ?
+	&text('select_device', "SD-Card", "$1") :
        $_[0] =~ /rd\/c(\d+)d(\d+)p(\d+)$/ ?
 	&text('select_mpart', "$1", "$2", "$3") :
        $_[0] =~ /ida\/c(\d+)d(\d+)p(\d+)$/ ?
@@ -2422,5 +2531,43 @@ if ($path =~ /^\\\\([^\\]+)\\([^\\]+)(\\.*)?/) {
 return $path;
 }
 
+# get_filesystem_label(device, mount, type)
+# Returns the label if there is one for this filesystem, or undef
+sub get_filesystem_label
+{
+local @p = @_;
+if (defined($get_filesystem_label_cache{$p[0]})) {
+	return $get_filesystem_label_cache{$p[0]};
+	}
+local $rv = "";
+if ($p[2] =~ /^ext\d+$/ && $has_e2label ||
+    $p[2] eq "xfs" && $has_xfs_db ||
+    $p[2] eq "reiserfs" && $has_reiserfstune) {
+	local $label;
+	if ($p[2] eq "xfs") {
+		local $out = &backquote_command(
+			"xfs_db -x -p xfs_admin -c label -r ".
+			quotemeta($p[0])." 2>&1", 1);
+		$label = $1 if ($out =~ /label\s*=\s*"(.*)"/ &&
+				$1 ne '(null)');
+		}
+	elsif ($p[2] eq "reiserfs") {
+		local $out = &backquote_command(
+			"reiserfstune ".quotemeta($p[0])." 2>&1");
+		if ($out =~ /LABEL:\s*(\S+)/) {
+			$label = $1;
+			}
+		}
+	else {
+		$label = &backquote_command(
+			"e2label ".quotemeta($p[0])." 2>&1", 1);
+		chop($label);
+		}
+	$rv = $label if (!$?);
+	}
+$get_filesystem_label_cache{$p[0]} = $rv;
+return $rv;
+}
+
 1;
 
